import numpy as np
from einsum_pipe import einsum_pipe


def einsum_pipe_simple(*args):
    subs = [arg for arg in args if isinstance(arg, (str, list, tuple))]
    ops = [arg for arg in args if not isinstance(arg, (str, list, tuple))]

    state: np.ndarray = ops.pop(0)
    for sub in subs:
        if isinstance(sub, str):
            extra_state = [ops.pop(0) for _ in range(sub.count(','))]
            state = np.einsum(sub, state, *extra_state)
        else:
            state = np.reshape(state, sub)

    return state


def test_single_arg_no_reshape():
    A = np.random.rand(10, 20, 30)
    args = [
        'abc->bca',
        'abc->ab',
        'ab->a',
        A
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_single_arg_with_reshape():
    A = np.random.rand(10, 20, 30)
    args = [
        'abc->bca',
        (10, 2, 5, 3, 2, 10),
        'abdfbg->afg',
        'abc->a',
        A
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_multi_args():
    A = np.random.rand(10, 20, 30)
    B = np.random.rand(10, 20, 30)
    args = [
        'abc,dec->bcad',
        (10, 2, 5, 3, 2, 10, 5, 2),
        'abcdbace->eacd',
        'abcd->acd',
        A, B
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_implicit_mode():
    A = np.random.rand(10, 20, 30)
    B = np.random.rand(10, 20, 30)
    args = [
        'abc,dec',
        'acbd',
        'cbdd',
        A, B
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_broadcasting():
    A = np.random.rand(10, 20, 30)
    B = np.random.rand(10, 20, 30)
    args = [
        'abc,...c->bca...',
        (10, 2, 5, 3, 2, 10, 5, 2, 20),
        'abcdbace...->eacd...',
        'abcd...->acd...',
        'acb...',
        'c...ba',
        A, B
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_unequal_broadcasting():
    A = np.random.rand(10, 20, 20)
    B = np.random.rand(10, 20, 30)
    args = [
        'ab...,...c->bca...',
        A, B
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_implicit_array_creation():
    A = np.random.rand(10, 20, 30)
    args = [
        'abc,->bca',
        'abc->a',
        A, 10
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_multistage_args():
    A = np.random.rand(10, 20, 30)
    B = np.random.rand(10, 20, 30)
    args = [
        'bac',
        'abc,dec->bcad',
        (10, 2, 5, 3, 2, 10, 5, 2),
        'abcdbace,->eacd',
        'abcd->acd',
        A, B, 10
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))


def test_implicit_broadcast():
    A = np.random.rand(10)
    B = np.random.rand(10)
    args = [
        '...,...',
        A, B
    ]
    assert np.allclose(einsum_pipe(*args), einsum_pipe_simple(*args))



import numpy as np
from einsum_pipe import einsum_pipe, ops


def test_trace():
    A = np.random.rand(10, 10, 10, 10)
    assert np.allclose(einsum_pipe(ops.trace(-3, 3), A),
                       np.trace(A, axis1=-3, axis2=3))


def test_diagonal():
    A = np.random.rand(10, 20, 10, 30)
    assert np.allclose(einsum_pipe(ops.diagonal(0, 2), A),
                       np.diagonal(A, axis1=0, axis2=2))
    assert np.allclose(einsum_pipe(ops.diagonal(-4, 2), A),
                       np.diagonal(A, axis1=-4, axis2=2))


def test_sum():
    A = np.random.rand(10, 20, 10, 30)
    assert np.allclose(einsum_pipe(ops.sum((3, -3, 0)), A),
                       np.sum(A, (3, -3, 0)))


def test_transpose():
    A = np.random.rand(10, 20, 10, 30)
    assert np.allclose(einsum_pipe(ops.transpose((3, -3, 0, -2)), A),
                       np.transpose(A, (3, -3, 0, -2)))


def test_inner():
    A = np.random.rand(2, 3, 30)
    B = np.random.rand(30)
    assert np.allclose(einsum_pipe(ops.inner(), A, B),
                       np.inner(A, B))


def test_outer():
    A = np.random.rand(2, 3, 30)
    B = np.random.rand(2, 3, 30)
    assert np.allclose(einsum_pipe(ops.outer(), A, B),
                       np.outer(A, B))



from typing import Dict, Generic, Iterator, TypeVar
from collections.abc import MutableMapping


K = TypeVar('K')
V = TypeVar('V')


class _BiDict(MutableMapping, Generic[K, V]):
    """A custom dictionary that keeps track of the inverse mapping from values to keys. Values must be unique too
    """

    def __init__(self):
        self.store: Dict[K, V] = {}
        self.inverse: Dict[V, K] = {}

    def __getitem__(self, key: K) -> V:
        return self.store[key]

    def __setitem__(self, key: K, value: V) -> None:
        if key in self:
            del self.inverse[self[key]]
        self.inverse[value] = key
        self.store[key] = value

    def __delitem__(self, key: K) -> None:
        del self.inverse[self[key]]
        del self.store[key]

    def values(self):
        return self.inverse.keys()

    def __iter__(self) -> Iterator[K]:
        return iter(self.store)

    def __len__(self):
        return len(self.store)



from typing import Any, Callable, List, Tuple, Union
import math


Subs = Union[str, Callable[[List[List[int]]], Union[str, List, Tuple]]]


def _build_base_subs(axis1: int, axis2: int) -> List[str]:
    # Unfortunately this function isn't universal: if the negative indices overlap with the
    # positive indices, i.e. an array with 5 dims and a reference to axis 4 and -5 produces
    # an invalid output. As such, it's only safe to use this if both axes are positive or both is negative.
    axis1, axis2 = sorted([axis1, axis2])
    subs = [chr(ord('a') + i) for i in range(axis2 + 1)] + ['...'] + \
        [chr(ord('a') + max(axis2 + 1, 0) + i - axis1)
         for i in range(axis1, 0)]
    return subs


def _same_sign(a: int, b: int) -> bool:
    return (a >= 0 and b >= 0) or (a < 0 and b < 0)


def _pop_many(data: List[Any], *axes: int) -> None:
    axes = tuple(sorted(ax if ax >= 0 else (len(data) + ax) for ax in axes))
    for ax in axes[::-1]:
        data.pop(ax)


def trace(axis1=0, axis2=1) -> Subs:
    if _same_sign(axis1, axis2):
        subs = _build_base_subs(axis1, axis2)
        subs[axis2] = subs[axis1]
        return ''.join(subs)
    else:
        def inner(input_shapes: List[List[int]]) -> str:
            subs = [chr(ord('a') + i) for i in range(len(input_shapes[0]))]
            subs[axis2] = subs[axis1]
            return ''.join(subs)
        return inner


def diagonal(axis1=0, axis2=1) -> Subs:
    axis1, axis2 = sorted([axis1, axis2])

    def from_subs(subs):
        subs[axis2] = subs[axis1]
        in_subs = ''.join(subs)
        sub = subs[axis1]
        _pop_many(subs, axis1, axis2)
        out_subs = ''.join(subs) + sub
        return in_subs + '->' + out_subs

    if _same_sign(axis1, axis2):
        subs = _build_base_subs(axis1, axis2)
        return from_subs(subs)
    else:
        def inner(input_shapes: List[List[int]]) -> str:
            subs = [chr(ord('a') + i) for i in range(len(input_shapes[0]))]
            return from_subs(subs)
        return inner


def diag() -> Subs:
    return diagonal()


def sum(axis: Union[None, int, List[int], Tuple[int, ...]] = None) -> Subs:
    if axis is None:
        return '...->'
    else:
        if isinstance(axis, int):
            axis = [axis]
        axis = sorted(axis)

        def from_subs(subs):
            in_subs = ''.join(subs)
            for ax in axis:
                subs[ax] = None  # type: ignore
            subs = [sub for sub in subs if sub is not None]
            out_subs = ''.join(subs)
            return in_subs + '->' + out_subs

        if _same_sign(axis[0], axis[-1]):
            subs = _build_base_subs(axis[0], axis[-1])
            return from_subs(subs)
        else:
            def inner(input_shapes: List[List[int]]) -> str:
                subs = [chr(ord('a') + i) for i in range(len(input_shapes[0]))]
                return from_subs(subs)
            return inner


def transpose(axes: Union[None, List[int], Tuple[int, ...]]) -> Subs:
    if axes is None:
        def inner(input_shapes: List[List[int]]) -> str:
            in_subs = ''.join([chr(ord('a') + i)
                              for i in range(len(input_shapes[0]))])
            return in_subs[::-1]
        return inner
    else:
        in_subs = [chr(ord('a') + i) for i in range(len(axes))]
        out_subs = [in_subs[a] for a in axes]
        return ''.join(in_subs) + '->' + ''.join(out_subs)


def inner() -> Subs:
    return '...a,...a'


def outer() -> Subs:
    def inner(input_shapes: List[List[int]]):
        out_shape = [math.prod(input_shapes[0]), math.prod(input_shapes[1])]
        in_subs = [chr(ord('a') + i)
                   for i in range(len(input_shapes[0]) + len(input_shapes[1]))]
        in_subs.insert(len(input_shapes[0]), ',')
        in_subs = ''.join(in_subs)
        return in_subs, out_shape
    return inner
